import math
from enum import Enum

from . import rotation
from .geometry_utils import test_line_segment_cross_dateline
from .resample_line import sample_next_point
from .spherical_triangle import get_third_vertex


class Polarity(Enum):
    UNKNOWN = 0
    RIGHT = 1
    LEFT = 2


def get_triangle(point_a, point_b, base_length, height, polarity):
    """return a triangle

    :params point_a:  {"lon": lon, "lat": lat} in radian
    :params point_b:  {"lon": lon, "lat": lat} in radian
    :params base_length: triangle base length
    :params polarity: subduction polarity
    """

    assert point_a != point_b  # the two points cannot be the same

    axis, angle = rotation.find_axis_and_angle(
        (point_a["lat"], point_a["lon"]),
        (point_b["lat"], point_b["lon"]),
    )

    if not base_length:
        lat = point_b["lat"]
        lon = point_b["lon"]
    else:
        lat, lon = rotation.rotate((point_a["lat"], point_a["lon"]), axis, base_length)
    _height = height
    if polarity == Polarity.RIGHT:
        _height = -height
    point_c = get_third_vertex(
        point_a,
        {"lon": lon, "lat": lat},
        _height,
    )

    return [
        point_a,
        {"lon": lon, "lat": lat},
        point_c,
    ]


def get_subduction_teeth(
    lons: list[float],
    lats: list[float],
    base_length: float = 0.03,
    spacing: float = 0.01,
    height: float = 0.03,
    polarity: Polarity = Polarity.UNKNOWN,
    geojson: bool = False,
    return_degrees: bool = True,
    central_meridian: float = None,
):
    """return a list of triangle polygons for given polyline

    :params lons: a list of lons (in radian)
    :params lats: a list of lats (in radian)
    :params base_length: base length of triangle (in radian)
    :params spacing: space between triangles (in radian)
    :params height: triangle height (in radian)
    :params polarity: the polarity of this subduction zone (left or right)
    :params geojson: if set True, return data in geojson format
    :params return_degrees: if set True, return coordinates in degrees
    :params central_meridian: if set, omit triangles cross dateline (in radian)

    :returns: [[{'lon':lon, 'lat':lat}, {'lon':lon, 'lat':lat}, {'lon':lon, 'lat':lat}],...]
        or {
                "type": "GeometryCollection",
                "geometries":
                [
                    {
                        "type": "Polygon",
                        "coordinates":
                        [
                            [
                                [100.0, 0.0], [101.0, 0.0], [101.0, 1.0],
                            ]
                        ]
                    },
                    {
                        "type": "Polygon",
                        "coordinates":
                        [
                            [
                                [110.0, 10.0], [10.0, 10.0], [101.0, 10.0],
                            ]
                        ]
                    },
                    ...
                ]
            }
    """
    assert len(lons) == len(lats)
    assert len(lons) > 1
    if (
        max(lons) > math.pi
        or min(lons) < -math.pi
        or max(lats) > math.pi
        or min(lats) < -math.pi
    ):
        raise Exception("The coordinates need to be in radians.")

    if geojson:
        return_degrees = True

    first_vertex = {"lon": lons[0], "lat": lats[0]}

    pp, ll = sample_next_point(list(zip(lats, lons)), base_length, strict=True)
    triangles = []
    while pp:
        second_vertex = {"lon": pp[1], "lat": pp[0]}
        triangles.append(
            get_triangle(
                first_vertex,
                second_vertex,
                None,  # base_length
                height,
                polarity,
            )
        )
        if polarity == Polarity.UNKNOWN:
            triangles.append(
                get_triangle(
                    first_vertex,
                    second_vertex,
                    None,  # base_length
                    height,
                    Polarity.LEFT,
                )
            )
            triangles.append(
                get_triangle(
                    first_vertex,
                    second_vertex,
                    None,  # base_length
                    height,
                    Polarity.RIGHT,
                )
            )

        pp, ll = sample_next_point(ll, spacing, strict=True)
        if not pp:
            break
        first_vertex = {"lon": pp[1], "lat": pp[0]}
        pp, ll = sample_next_point(ll, base_length, strict=True)

    _triangles = []
    if central_meridian is not None:
        for triangle in triangles:
            if (
                test_line_segment_cross_dateline(
                    triangle[0]["lon"], triangle[1]["lon"], central_meridian
                )
                or test_line_segment_cross_dateline(
                    triangle[1]["lon"], triangle[2]["lon"], central_meridian
                )
                or test_line_segment_cross_dateline(
                    triangle[0]["lon"], triangle[2]["lon"], central_meridian
                )
            ):
                continue
            else:
                _triangles.append(triangle)
        triangles = _triangles

    if geojson:
        polygons = []
        for triangle in triangles:
            polygon = {"type": "Polygon"}
            if return_degrees:
                polygon["coordinates"] = [
                    [
                        [
                            math.degrees(triangle[0]["lon"]),
                            math.degrees(triangle[0]["lat"]),
                        ],
                        [
                            math.degrees(triangle[1]["lon"]),
                            math.degrees(triangle[1]["lat"]),
                        ],
                        [
                            math.degrees(triangle[2]["lon"]),
                            math.degrees(triangle[2]["lat"]),
                        ],
                        [
                            math.degrees(triangle[0]["lon"]),
                            math.degrees(triangle[0]["lat"]),
                        ],
                    ]
                ]
            else:
                polygon["coordinates"] = [
                    [
                        [triangle[0]["lon"], triangle[0]["lat"]],
                        [triangle[1]["lon"], triangle[1]["lat"]],
                        [triangle[2]["lon"], triangle[2]["lat"]],
                        [triangle[0]["lon"], triangle[0]["lat"]],
                    ]
                ]
            polygons.append(polygon)
        return polygons

    if return_degrees:
        triangles_in_degrees = []
        for triangle in triangles:
            triangle_in_degrees = [
                {
                    "lon": math.degrees(triangle[0]["lon"]),
                    "lat": math.degrees(triangle[0]["lat"]),
                },
                {
                    "lon": math.degrees(triangle[1]["lon"]),
                    "lat": math.degrees(triangle[1]["lat"]),
                },
                {
                    "lon": math.degrees(triangle[2]["lon"]),
                    "lat": math.degrees(triangle[2]["lat"]),
                },
            ]
            triangles_in_degrees.append(triangle_in_degrees)
        return triangles_in_degrees
    else:
        return triangles


def get_subduction_teeth_in_degrees(
    lons: list[float],
    lats: list[float],
    base_length: float = 1,
    spacing: float = 0.5,
    height: float = 1,
    polarity: Polarity = Polarity.UNKNOWN,
    geojson: bool = False,
    central_meridian: float = None,
):
    """same as get_subduction_teeth, just everything is in degrees"""
    if central_meridian is not None:
        central_meridian = math.radians(central_meridian)
    return get_subduction_teeth(
        [math.radians(lon) for lon in lons],
        [math.radians(lat) for lat in lats],
        base_length=math.radians(base_length),
        spacing=math.radians(spacing),
        height=math.radians(height),
        polarity=polarity,
        geojson=geojson,
        return_degrees=True,
        central_meridian=central_meridian,
    )



import os
from functools import wraps

from dotenv import load_dotenv

script_path = os.path.dirname(os.path.realpath(__file__))
cwd = os.getcwd()

username = None
passwd = None
server_url = None
proxy = ""


def auth(func):
    """decorator to get server authentication info

    :param func: the user's function which this decorator will wrap around.

    """

    @wraps(func)
    def inner(*args, **kwargs):
        """the wrapper function

        :param *args: allow the inner function to accept multiple positional arguments
        :param **kwargs: allow the inner function to accept multiple keyword (or named) arguments.

        """
        global username
        global passwd
        global server_url
        global proxy
        if not username or not passwd or not server_url:
            username, passwd, server_url, proxy = get_cfg()
        return func(*args, **kwargs)

    return inner


def get_env():
    """Get username, password, url and proxy from environment."""
    return (
        os.environ.get("GWS_USERNAME"),
        os.environ.get("GWS_PASSWORD"),
        os.environ.get("GWS_URL"),
        os.environ.get("GWS_PROXY"),
    )


def get_cfg():
    """Get the server configuration, such as username, password, gplates web service URL, proxy
    either from environment variables or .env file.

    """
    # first, try to get the info from environment variables
    username, passwd, server_url, proxy = get_env()
    # if cound not get all info from environment variables
    # try to load .env file
    if not all([username, passwd, server_url]):
        # load environment variables from .env.
        load_dotenv(f"{cwd}/.env")
        username, passwd, server_url, proxy = get_env()
        # still failed? inform caller something is wrong
        if not server_url:
            server_url = "https://gws.gplates.org"
    return username, passwd, server_url, proxy



import math

from . import rotation


def degree_to_radian(point):
    return [math.radians(point[0]), math.radians(point[1])]


def radian_to_degree(point):
    return [math.degrees(point[0]), math.degrees(point[1])]


def sample_next_point(line, distance, strict=False, accum_length=0):
    """recursive function to sample point one by one
    :param line: a list of (lat, lon) in radians
    :param distance: the distance to the next point
    :param strict: in strict mode, we guarantee the distance. sacrifice the end point if necessary
    :accum_length: parameter to pass down the accumulated length for this recursive function
    :returns: (next point, new line)
        next point (lat,lon) in radians. None, if no more point
        new line -- the line after taking out this point
    """
    if len(list(line)) < 2:
        return None, []

    start_point = line[0]
    second_point = line[1]

    dd = rotation.radian_distance(start_point, second_point)

    if math.isclose(dd + accum_length, distance):
        return second_point, line[1:]
    elif (dd + accum_length) > distance:
        pp = rotation.sample_between_two_points(
            start_point, second_point, [distance - accum_length]
        )[0]
        assert pp
        return pp, [pp] + line[1:]
    else:
        if len(line[1:]) > 1:
            return sample_next_point(
                line[1:], distance, strict=strict, accum_length=dd + accum_length
            )
        else:
            if strict:
                # in strict mode, we guarantee the distance. It means we will sacrifice the end point if necessary
                return None, []
            else:
                return second_point, line[1:]  # reached the last point


def resample_line(line, distance, strict=False):
    """resample a line.
    :param line: a list of (lat, lon) in radians
    :param distance: the distance of the adjacent points
    :param strict: in strict mode, we guarantee the distance. sacrifice the end point if necessary
        in unstrict mode, we always keep the end point
    :returns: the new line
        a list of (lat, lon) in radians. If the input line is shorter than distance, in strict mode, return []
    """
    new_line = [line[0]]

    ll = line

    while True:
        pp, ll = sample_next_point(ll, distance, strict=strict)
        if pp:
            new_line.append(pp)
        else:
            break

    if len(new_line) > 1:
        return new_line
    else:
        return []


if __name__ == "__main__":
    pass



#!/usr/bin/env python3

import sys
from pathlib import Path

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

OUTPUT_DIR = "output"
Path(OUTPUT_DIR).mkdir(parents=True, exist_ok=True)

from gplates_ws_proxy import coastlines, paleoearth

# dev test
# export GWS_URL=http://localhost:18000/
# micromamba run -n gplates-ws-example ./plot_paleo_labels.py


def main(show=True):
    time = 100
    labels = paleoearth.get_paleo_labels(time=time, model="Merdith2021")
    coastlines_shapely = coastlines.get_paleo_coastlines(
        time=time, model="Merdith2021", format="shapely"
    )

    fig = plt.figure(figsize=(12, 6), dpi=120)
    ax = plt.axes(projection=ccrs.Robinson())

    ax.set_global()
    ax.gridlines()

    ax.add_geometries(
        coastlines_shapely,
        crs=ccrs.PlateCarree(),
        facecolor="lightgrey",
        edgecolor="lightgrey",
        alpha=1,
    )

    for name, lon, lat in zip(labels["names"], labels["lons"], labels["lats"]):
        plt.text(
            lon,
            lat,
            name,
            va="baseline",
            family="monospace",
            weight="bold",
            color="red",
            transform=ccrs.PlateCarree(),
        )

    plt.plot(labels["lons"], labels["lats"], "o", transform=ccrs.PlateCarree())
    plt.title(f"{time} Ma", fontsize=20)

    if show:
        plt.show()
    else:
        output_file = f"{OUTPUT_DIR}/{Path(__file__).stem}.png"
        fig.savefig(output_file, dpi=120, bbox_inches="tight")
        print(f"Done! The {output_file} has been saved.")
        plt.close(fig)


if __name__ == "__main__":
    if len(sys.argv) == 2 and sys.argv[1] == "save":
        main(show=False)
    else:
        main(show=True)



import requests

from . import _auth as a
from ._auth import auth


@auth
def get_paleo_labels(time: float, model: str = "MULLER2022"):
    """get a list of paleo-labels

    :returns: a dict object, {"names": names, "lons": lons, "lats": lats}
    """

    url = f"{a.server_url}/earth/get_labels?time={time}&model={model}"
    r = requests.get(url)
    if r.status_code != 200:
        raise Exception(f"Failed to get paleo-labels {r.status_code}! -- {url}")
    labels = r.json()
    names = []
    lons = []
    lats = []
    for label in labels:
        names.append(label[0])
        lats.append(label[1])
        lons.append(label[2])
    return {"names": names, "lons": lons, "lats": lats}


@auth
def get_paleo_cities(time: float, model: str = "MULLER2022"):
    """get a list of paleo-cities

    :returns: a dict object, {"names": names, "lons": lons, "lats": lats}
    """

    url = f"{a.server_url}/earth/get_cities?time={time}&model={model}"
    r = requests.get(url)
    if r.status_code != 200:
        raise Exception(f"Failed to get paleo-cities {r.status_code}! -- {url}")

    return r.json()



import requests
from shapely.geometry import LineString, shape

from . import _auth as a
from ._auth import auth
from .utils import geojson_to_shapely


class Topology:
    def __init__(self, model_name, time):
        self.plate_polygons = None
        self.plate_polygons_lines = None
        self.plate_boundaries = None
        self.model_name = model_name
        self.time = time

    def get_plate_polygons(self, return_format="geojson", as_lines=False):
        if not as_lines:
            if self.plate_polygons is None:
                self.plate_polygons = get_topological_plate_polygons(
                    self.model_name, self.time
                )
            if return_format.lower() == "geojson":
                return self.plate_polygons
            else:
                return geojson_to_shapely(self.plate_polygons)
        else:
            if self.plate_polygons_lines is None:
                self.plate_polygons_lines = get_topological_plate_polygons(
                    self.model_name, self.time, as_lines=True
                )
            if return_format.lower() == "geojson":
                return self.plate_polygons_lines
            else:
                return geojson_to_shapely(self.plate_polygons_lines)

    def get_plate_boundaries(self, return_format="geojson"):
        if self.plate_boundaries is None:
            self.plate_boundaries = get_topological_plate_boundaries(
                self.model_name, self.time
            )
        if return_format.lower() == "geojson":
            return self.plate_boundaries
        else:
            return geojson_to_shapely(self.plate_boundaries)

    def get_plate_polygon_names(self):
        names = []
        if self.plate_polygons is None:
            self.plate_polygons = get_topological_plate_polygons(
                self.model_name, self.time
            )
        for feature in self.plate_polygons["features"]:
            if "name" in feature["properties"] and feature["properties"]["name"]:
                names.append(feature["properties"]["name"])
        return list(set(names))

    def get_plate_boundary_types(self):
        types = []
        if self.plate_boundaries is None:
            self.plate_boundaries = get_topological_plate_boundaries(
                self.model_name, self.time
            )
        for feature in self.plate_boundaries["features"]:
            if "type" in feature["properties"]:
                types.append(feature["properties"]["type"])
        return list(set(types))

    def get_features(self, name, return_format="geojson"):
        ret = {"type": "FeatureCollection", "features": []}

        if self.plate_boundaries is None:
            self.plate_boundaries = get_topological_plate_boundaries(
                self.model_name, self.time
            )
        for feature in self.plate_boundaries["features"]:
            if (
                "type" in feature["properties"]
                and feature["properties"]["type"] == name
            ):
                ret["features"].append(feature)

        if return_format.lower() == "geojson":
            return ret
        else:
            return geojson_to_shapely(ret)


@auth
def get_topological_plate_boundaries(model, time):
    """return topological plate boundaries

    :params model: model name
    :params time: reconstruction time

    :returns: geojson
    """
    headers = {
        "Accept": "application/json",
    }

    params = {"time": time, "model": model}

    ret = requests.get(
        a.server_url + "/topology/plate_boundaries/",
        params=params,
        verify=True,
        headers=headers,
        proxies={"http": a.proxy},
    )

    if ret.status_code in [200, 201]:
        return ret.json()


@auth
def get_topological_plate_polygons(model, time, as_lines=False):
    """return topological plate polygons

    :params model: model name
    :params time: reconstruction time

    :returns: geojson
    """
    headers = {
        "Accept": "application/json",
    }

    params = {"time": time, "model": model}
    if as_lines:
        params["as_lines"] = True

    ret = requests.get(
        a.server_url + "/topology/plate_polygons/",
        params=params,
        verify=True,
        headers=headers,
        proxies={"http": a.proxy},
    )

    if ret.status_code in [200, 201]:
        return ret.json()


@auth
def get_subduction_zones(model, time):
    """return subduction zones as geojson object"""
    r = requests.get(
        f"{a.server_url}/topology/get_subduction_zones?time={time}&model={model}"
    )
    if r.status_code != 200:
        raise Exception("Failed to get subduction zones.")
    return r.json()



import math


def normalize(v, tolerance=0.00001):
    mag2 = sum(n * n for n in v)

    assert not math.isclose(mag2, 0)  # the magnitude of vector cannot be zero.

    if abs(mag2 - 1.0) > tolerance:
        mag = math.sqrt(mag2)
        v = tuple(n / mag for n in v)
    return v


def quat_mult(q1, q2):
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return w, x, y, z


def quat_conjugate(q):
    q = normalize(q)
    w, x, y, z = q
    return (w, -x, -y, -z)


def quat_vec_mult(q1, v1):
    v1 = normalize(v1)
    q2 = (0.0,) + v1
    return quat_mult(quat_mult(q1, q2), quat_conjugate(q1))[1:]


def axis_angle_to_quat(v, theta):
    v = normalize(v)
    x, y, z = v
    theta /= 2
    w = math.cos(theta)
    x = x * math.sin(theta)
    y = y * math.sin(theta)
    z = z * math.sin(theta)
    return w, x, y, z


def quat_to_axis_angle(quat):
    w, v = quat[0], quat[1:]
    theta = math.acos(w) * 2.0
    return normalize(v), theta


def lat_lon_to_cart(lat, lon):
    x = math.cos(lat) * math.cos(lon)
    y = math.cos(lat) * math.sin(lon)
    z = math.sin(lat)
    return x, y, z


def cart_to_lat_lon(x, y, z):
    lat = math.asin(z)
    lon = math.atan2(y, x)
    return lat, lon


def test_rotate(point, axis, angle):
    """test rotate point

    :param point: point to be rotated (in lat,lon)
    :param axis: rotation axis (in lat,lon)
    :param angle: roation angle (in degrees)

    :returns: new coordinates of the point (in lat,lon)
    """
    v = lat_lon_to_cart(math.radians(point[0]), math.radians(point[1]))
    axis = lat_lon_to_cart(math.radians(axis[0]), math.radians(axis[1]))
    quat = axis_angle_to_quat(axis, math.radians(angle))
    ret = quat_vec_mult(quat, v)
    ret_lat_lon = cart_to_lat_lon(ret[0], ret[1], ret[2])
    return math.degrees(ret_lat_lon[0]), math.degrees(ret_lat_lon[1])


if __name__ == "__main__":
    print(test_rotate([12, 134], [90, 0], 34))



import math

from . import quaternions

#
# a collection of rotation related functions
#


def cross(a, b):
    """cross product of two vectors defined only in three-dimensional space
    https://www.mathsisfun.com/algebra/vectors-cross-product.html
    https://en.wikipedia.org/wiki/Cross_product#Computing

    :param a: 3D vector, such as [1,2,3]
    :param b: 3D vector, such as [234]

    :returns: cross product 3D vector
    """

    c = [
        a[1] * b[2] - a[2] * b[1],
        a[2] * b[0] - a[0] * b[2],
        a[0] * b[1] - a[1] * b[0],
    ]
    return c


def dot(a, b):
    """dot product of two vectors
    https://www.mathsisfun.com/algebra/vectors-dot-product.html

    :param a: a vector, such as [1,2,3,4,5] (must be the same length with b)
    :param b: a vector, such as [6,7,8,9,10] (must be the same length with a)

    :returns: dot product (a number)
    """
    return sum(i * j for i, j in zip(a, b))


def find_axis_and_angle(point_a, point_b):
    """given two points point_a and point_b in (lat, lon) format in radians,
    return axis(lat, lon) and angle in radians

    :param point_a: point a (lat, lon) in radians
    :param point_b: point b (lat, lon) in radians

    :returns: axis(lat, lon) in radians; angle in radians
    """
    # get 3D vectors
    a_v = quaternions.lat_lon_to_cart(point_a[0], point_a[1])
    b_v = quaternions.lat_lon_to_cart(point_b[0], point_b[1])

    ab_cross = quaternions.normalize(cross(a_v, b_v))
    ab_dot = dot(a_v, b_v)

    axis = quaternions.cart_to_lat_lon(*ab_cross)  # in radians
    angle = math.acos(ab_dot)  # in radians

    return axis, angle


def rotate(point, axis, angle):
    """rotate a point by axis and angle

    :param point: (lat, lon) in radians
    :param axis: (lat, lon) in radians
    :param angle: in radians

    :returns: new point(lat, lon) in radians
    """
    v = quaternions.lat_lon_to_cart(point[0], point[1])
    axis = quaternions.lat_lon_to_cart(axis[0], axis[1])
    quat = quaternions.axis_angle_to_quat(axis, angle)
    ret = quaternions.quat_vec_mult(quat, v)
    ret_lat_lon = quaternions.cart_to_lat_lon(ret[0], ret[1], ret[2])
    return (ret_lat_lon[0], ret_lat_lon[1])


def interp_two_points(point_a, point_b, num=10):
    """interpolate between two points(in radians) along the great circle
    return a list of interpolated points (lat, lon) in radians

    :param point_a: point a (lat, lon) in radians
    :param point_b: point b (lat, lon) in radians
    :param num: the number of segments

    :returns: a list of interpolated points (lat, lon) in radians
        the return list has num+1 points(num segments)
    """
    # make the two input points are not the same
    assert not (
        math.isclose(point_a[0], point_b[0]) and math.isclose(point_a[1], point_b[1])
    )
    ret = [point_a]
    axis, angle = find_axis_and_angle(point_a, point_b)
    for angle_ in [(angle / num) * i for i in range(1, num + 1)]:
        lat, lon = rotate(point_a, axis, angle_)
        ret.append((lat, lon))

    return ret


def sample_between_two_points(point_a, point_b, distances_to_point_a):
    """sample points between two points along the great circle according to a list of distances to point a

    :param point_a: point a (lat, lon) in radians
    :param point_b: point b (lat, lon) in radians
    :param distances_to_point_a: a list of distances(in radians) to point a.
        the distance must less than the distance between point a and point b.
        invalid distances will produce Nones in the return list

    :returns: a list of sampled points (lat, lon) in radians.
        The return list has the same length with the list of distances_to_point_a.
        point_a and point_b will not be included.

    """
    # make the two input points are not the same
    assert not (
        math.isclose(point_a[0], point_b[0]) and math.isclose(point_a[1], point_b[1])
    )
    ret = []
    axis, angle = find_axis_and_angle(point_a, point_b)
    for distance in distances_to_point_a:
        if distance < 0 or distance > angle:
            ret.append(None)
        else:
            lat, lon = rotate(point_a, axis, distance)
            ret.append((lat, lon))

    return ret


def distance(point_a, point_b, earth_radius=6378.14):
    """calculate the distance between to points (lat, lon) in radians
    return the distance in km

    :param point_a: point a (lat, lon) in radians
    :param point_b: point b (lat, lon) in radians
    :param earth_radius: earth radius

    :returns: distance in km
    """
    if math.isclose(point_a[0], point_b[0]) and math.isclose(point_a[1], point_b[1]):
        return 0
    _, angle = find_axis_and_angle(point_a, point_b)
    return radian_to_km(angle, earth_radius)


def radian_distance(point_a, point_b):
    """calculate the distance between to points (lat, lon) in radians
    return the distance in radians

    :param point_a: point a (lat, lon) in radians
    :param point_b: point b (lat, lon) in radians

    :returns: distance in radians
    """
    if math.isclose(point_a[0], point_b[0]) and math.isclose(point_a[1], point_b[1]):
        return 0
    _, angle = find_axis_and_angle(point_a, point_b)
    return angle


def km_to_radian(km, earth_radius=6378.14):
    return km / earth_radius


def radian_to_km(radian, earth_radius=6378.14):
    return radian * earth_radius


if __name__ == "__main__":
    # test
    # 1. find axis and angle between two points a and b
    # 2. use the axis and angle to rotate point a
    # 3. check if the rotation ends up at point b
    point_a = tuple((math.radians(i) for i in (45, 20)))  # lat, lon
    point_b = tuple((math.radians(i) for i in (-45, 120)))

    axis, angle = find_axis_and_angle(point_a, point_b)

    print(axis, angle)

    point_b_new = rotate(point_a, axis, angle)

    # check if the two points are the same
    print(tuple(round(math.degrees(i), 2) for i in point_b_new))
    print(tuple(round(math.degrees(i), 2) for i in point_b))

    # intepolate between point a and b
    points_in_between = interp_two_points(point_a, point_b)
    with open("interpolation.gmt", "w+") as f:
        f.write(f"{math.degrees(axis[1]):.2f} {math.degrees(axis[0]):.2f}\n")
        f.write(f"{math.degrees(axis[0]):.2f} {math.degrees(axis[1]):.2f}\n")
        for p in points_in_between:
            f.write(f"{math.degrees(p[1]):.2f} {math.degrees(p[0]):.2f}\n")

    print(distance(point_a, point_b))
    print("test finished!")



import sys

sys.path.insert(0, "../examples/")
import plot_paleo_coastlines
import plot_paleo_labels
import plot_subduction_zones
import plot_topological_features
import plot_topological_plate_boundaries
import plot_topological_plate_polygons
import reconstruct_shapely_points


def test():
    reconstruct_shapely_points.main(show=False)
    plot_topological_plate_polygons.main(show=False)
    plot_topological_plate_boundaries.main(show=False)
    plot_topological_features.main(show=False)
    plot_subduction_zones.main(show=False)
    plot_paleo_labels.main(show=False)
    plot_paleo_coastlines.main(show=False)



from utils import logger

from gplates_ws_proxy import PlateModel, utils

model = PlateModel("Muller2019")
time = 10
topology_10 = model.get_topology(time)

logger.info(utils.get_cfg())


def test_plate_polygon_names():
    msg = []
    for name in topology_10.get_plate_polygon_names():
        msg.append(name)
    logger.info(msg)


def test_plate_boundaries_types():
    msg = []
    for type in topology_10.get_plate_boundary_types():
        msg.append(type)
    logger.info(msg)


if __name__ == "__main__":
    test_plate_polygon_names()
    test_plate_boundaries_types()



from shapely.geometry import MultiPolygon, Polygon, shape

from . import _auth as a
from ._auth import auth


def geojson_to_shapely(json_data):
    ret = []
    for feature in json_data["features"]:
        s = shape(feature["geometry"])
        if isinstance(s, Polygon) or isinstance(s, MultiPolygon):
            ret.append(s.buffer(0))
        else:
            ret.append(s)

    return ret


@auth
def get_cfg():
    return {
        "username": a.username,
        "passwd": a.passwd,
        "server_url": a.server_url,
        "proxy": a.proxy,
    }



from utils import logger

import gplates_ws_proxy


def test_coastlines():
    r = gplates_ws_proxy.get_paleo_coastlines(100)
    print(r)

    r = gplates_ws_proxy.get_paleo_coastlines(100, format="shapely")
    print(r)



#!/usr/bin/env python3
from pathlib import Path

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import moviepy.editor as mpy

from gplates_ws_proxy import coastlines

OUTPUT_DIR = "output"
Path(OUTPUT_DIR).mkdir(parents=True, exist_ok=True)


def main():
    for time in range(0, 141, 10):
        get_basemap_with_coastlines(time=time)
        plt.title(f"{time} Ma")
        save_fig(f"paleo-coastlines-{time}-Ma.png")

    frame_list = [
        f"{OUTPUT_DIR}/paleo-coastlines-{time}-Ma.png" for time in range(0, 141, 10)
    ]
    # print(frame_list)
    clip = mpy.ImageSequenceClip(frame_list, fps=4)

    clip.write_videofile(
        f"{OUTPUT_DIR}/coastlines.mp4",
        codec="libx264",
        # audio_codec='aac',
        ffmpeg_params=["-s", "1440x720", "-pix_fmt", "yuv420p"],
    )  # LOOK HERE!!!! give image size here(the numbers must divide by 2)
    print("video has been created!")


def get_basemap_with_coastlines(model="Muller2019", crs=ccrs.Robinson(), time=140):
    coastlines_shapely = coastlines.get_paleo_coastlines(
        time=time, model=model, format="shapely", min_area=500
    )

    fig = plt.figure(figsize=(12, 6), dpi=120)
    ax = plt.axes(projection=crs)

    ax.set_global()
    ax.gridlines()

    ax.add_geometries(
        coastlines_shapely,
        crs=ccrs.PlateCarree(),
        facecolor="grey",
        edgecolor="red",
        alpha=0.5,
    )
    return ax


def save_fig(filename):
    output_file = f"{OUTPUT_DIR}/{Path(filename).stem}.png"
    plt.gcf().savefig(output_file, dpi=120, bbox_inches="tight")  # transparent=True)
    print(f"Done! The {output_file} has been saved.")
    plt.close(plt.gcf())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import io
import sys
from pathlib import Path

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

from gplates_ws_proxy import PlateModel

OUTPUT_DIR = "output"
Path(OUTPUT_DIR).mkdir(parents=True, exist_ok=True)

# dev test
# export GWS_URL=http://localhost:18000/
# micromamba run -n gplates-ws-example ./plot_paleo_coastlines.py


def main(show=True):
    fig = plt.figure(figsize=(12, 6), dpi=120)

    # plot 1
    ax1 = fig.add_subplot(221, projection=ccrs.PlateCarree())
    ax1.set_global()
    ax1.gridlines()

    model = PlateModel("Muller2022")
    coastlines_shapely = model.get_coastlines(
        time=100, extent=[-90, 90, -40, 40], min_area=10, format="shapely"
    )

    ax1.add_geometries(
        coastlines_shapely,
        crs=ccrs.PlateCarree(),
        facecolor="blue",
        edgecolor="none",
        alpha=0.5,
    )
    ax1.set_title(f"{100} Ma")

    # plot 2
    ax2 = fig.add_subplot(222, projection=ccrs.Robinson())
    ax2.set_extent([-100, 100, -50, 50], crs=ccrs.Geodetic())
    ax2.gridlines()

    model = PlateModel("Muller2019")
    time = 109
    coastlines_shapely = model.get_coastlines(time=time, min_area=500, format="shapely")

    ax2.add_geometries(
        coastlines_shapely,
        crs=ccrs.PlateCarree(),
        facecolor="green",
        edgecolor="none",
        alpha=0.5,
    )
    ax2.set_title(f"{time} Ma")

    # plot 3
    ax3 = fig.add_subplot(223, projection=ccrs.Mollweide())
    ax3.set_global()
    ax3.gridlines()

    model = PlateModel("Muller2022")
    coastlines_shapely = model.get_coastlines(
        time=10, min_area=70 * 1000, format="shapely"
    )

    ax3.add_geometries(
        coastlines_shapely,
        crs=ccrs.PlateCarree(),
        facecolor="green",
        edgecolor="none",
        alpha=0.5,
    )
    ax3.set_title(f"min area 70000 km2")

    # plot 4
    ax4 = fig.add_subplot(224, projection=ccrs.Robinson())
    ax4.set_global()
    ax4.gridlines()
    model = PlateModel("Muller2022")
    coastlines_png = model.get_coastlines(
        time=200, format="png", facecolor="lime", edgecolor="red"
    )
    # print(type(coastlines_png))
    data = plt.imread(io.BytesIO(coastlines_png), format="png")
    ax4.imshow(
        data, origin="upper", transform=ccrs.PlateCarree(), extent=[-180, 180, -90, 90]
    )
    ax4.set_title(f"200 Ma")

    fig.subplots_adjust(
        bottom=0.05, top=0.95, left=0.1, right=0.9, wspace=0.02, hspace=0.05
    )

    if show:
        plt.show()
    else:
        save_fig(__file__)


def save_fig(filename):
    output_file = f"{OUTPUT_DIR}/{Path(filename).stem}.png"
    plt.gcf().savefig(output_file, dpi=120, bbox_inches="tight")  # transparent=True)
    print(f"Done! The {output_file} has been saved.")
    plt.close(plt.gcf())


if __name__ == "__main__":
    if len(sys.argv) == 2 and sys.argv[1] == "save":
        main(show=False)
    else:
        main(show=True)



#!/usr/bin/env python3
import sys
from pathlib import Path

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from shapely.geometry import MultiLineString

from gplates_ws_proxy import PlateModel

OUTPUT_DIR = "output"
Path(OUTPUT_DIR).mkdir(parents=True, exist_ok=True)

# dev test
# export GWS_URL=http://localhost:18000/
# micromamba run -n gplates-ws-example ./plot_topological_features.py


def plot_lines(ax, lines, color="blue"):
    for line in lines:
        if isinstance(line, MultiLineString):
            for g in line.geoms:
                ax.plot(*g.xy, transform=ccrs.PlateCarree(), color=color)
        else:
            ax.plot(*line.xy, transform=ccrs.PlateCarree(), color=color)


def main(show=True):
    model = PlateModel("Muller2019")
    time = 10
    topology_10 = model.get_topology(time)

    mid_ccean_ridge = topology_10.get_features("MidOceanRidge", return_format="shapely")
    transform = topology_10.get_features("Transform", return_format="shapely")
    fault = topology_10.get_features("Fault", return_format="shapely")
    subduction = topology_10.get_features("SubductionZone", return_format="shapely")

    fig = plt.figure(figsize=(12, 6), dpi=120)
    ax = plt.axes(projection=ccrs.Robinson())

    ax.set_global()
    ax.gridlines()

    plot_lines(ax, mid_ccean_ridge, color="red")
    plot_lines(ax, transform, color="black")
    plot_lines(ax, fault, color="yellow")
    plot_lines(ax, subduction)

    plt.title(f"{time} Ma")
    if show:
        plt.show()
    else:
        save_fig(__file__)


def save_fig(filename):
    output_file = f"{OUTPUT_DIR}/{Path(filename).stem}.png"
    plt.gcf().savefig(output_file, dpi=120, bbox_inches="tight")  # transparent=True)
    print(f"Done! The {output_file} has been saved.")
    plt.close(plt.gcf())


if __name__ == "__main__":
    if len(sys.argv) == 2 and sys.argv[1] == "save":
        main(show=False)
    else:
        main(show=True)
